import rtlsdr
import numpy as N
import pylab, pickle, time, profile
import scipy.fftpack as fft
import scipy.signal
import database as DB

#import numpy.fft as fft

# TODO: Should sampling frequency be corrected for error in crystal?
# TODO: Track very short signals and paint a square around them in myFFT

# Container for parameters
class par: 
    pass
par.scanNfft, par.scanNavg, par.ppm = 1024, 4, 0
par.scanDrop = 0
par.corr = 1.0/(1.0+par.ppm*1.0e-6)
par.scanRes = 5 # Signals closer than this numer of steps in fft are considered the same

par.lookNfft, par.lookNavg = 1024*4, 1024
par.lookRate, par.lookSkip = 2.0e6, 4

def read_samples(radio,bytes):
    raw_data = radio.read_bytes(bytes*2)
    raw_data = N.ctypeslib.as_array(raw_data)
    iq = N.empty(bytes, 'complex')
    iq.real, iq.imag = raw_data[::2], raw_data[1::2]
    iq /= (255.0/2)
    iq -= (1 + 1j)
    return raw_data, iq
    
def calibrate(radio,par,freq,verbose):
    rate=radio.get_sample_rate()
    cfreq = freq+500e3
    radio.set_center_freq(cfreq)
    time.sleep(0.01)
    raw_data, data = read_samples(radio,par.lookNfft*par.lookNavg)

    F=N.reshape(data,(par.lookNavg,par.lookNfft))
    F=N.roll(N.abs(fft.fft(F[:,:]))**2,par.lookNfft/2)
    Fmean = N.mean(F,0)
    peaks, noise, noiseStd = peakDetect(Fmean)
    dF=(rate/par.lookNfft)
    closest, msnr = None, 0
    for fist, last, snr in peaks:
        ff=cfreq+(fist-par.lookNfft/2)*dF
        fl=cfreq+(last-par.lookNfft/2)*dF
        mfreq=(ff+fl)/2
        if snr>msnr and abs(mfreq-freq)/freq<300e-6:
            closest, msnr = mfreq, snr
    if closest!=None:
        par.corr = float(closest)/float(freq) 
        par.ppm = (1.0/par.corr-1.0)*1e6

    if verbose: 
        if closest!=None:
            print "Calibration PPM %3.1f"%par.ppm
        else:
            print "Failed to calibrate"

def look(radio,par,lfreq,verbose):
    rate=radio.get_sample_rate()
    freq=(lfreq-rate/4)
    if verbose: print "Looking at %4.2f for %4.2f"%(freq/1e6,lfreq/1e6)
    radio.set_center_freq(freq*par.corr)
    raw_data, data = read_samples(radio,par.lookNfft*par.lookNavg)

    F=N.reshape(data,(par.lookNavg,par.lookNfft))
    F=N.roll(N.abs(fft.fft(F[::par.lookSkip,:]))**2,par.lookNfft/2)
    Fmean = N.mean(F,0)
    # Assume that signals have narrow bandwith -> noise average 
    noise, noiseStd = N.mean(Fmean), N.std(Fmean)

    peaks = [peakDetect(F[ii,:],True)[0] for ii in range(len(F))]
    
    dF=(rate/par.lookNfft/par.corr)
    found = False
    for peak in peaks:
        for fist, last, snr in peak:
            ff=freq+(fist-par.lookNfft/2)*dF
            fl=freq+(last-par.lookNfft/2)*dF
            found = True

    fn = '%i'%(hash(freq+time.time()))+'.npy'
    if found:
        # Save
        N.save('SavedSignals/'+fn,raw_data)
    return found, freq, lfreq, fn 

def peek(radio,par,verbose,peekFreq):
    rate=radio.get_sample_rate()

    savePeaks = []
    for freq in peekFreq:
        if verbose: print "Freq : ",freq
        radio.set_center_freq(freq*par.corr)
        time.sleep(0.005)
        raw_data, data = read_samples(radio,par.scanNfft*par.scanNavg)
        data=N.reshape(data,(par.scanNavg,par.scanNfft))
        F=N.roll(N.abs(fft.fft(data[par.scanDrop:,:]))**2,par.scanNfft/2)
        Fmean = N.mean(F,0)

        peaks, noise, noiseStd = peakDetect(Fmean)

        peaksF = []
        for fist, last, snr in peaks:
            dF=(rate/par.scanNfft/par.corr)
            ff=freq+(fist-par.scanNfft/2)*dF
            fl=freq+(last-par.scanNfft/2)*dF
            if verbose:
                print "%3.2f MHz (FWHM %3.0f kHz) SNR: %3.0f"%((ff+fl)/2e6,(dF+fl-ff)/1e3,snr)
            peaksF+= [[(ff+fl)/2e6,(fl-ff)/1e6,snr]]

        if len(peaks)>20 and False:
            xx=(freq+(N.arange(par.scanNfft)-par.scanNfft/2)*dF)/1e6
            tmp=10*N.log(Fmean/noise)/N.log(10)
            pylab.figure(0)
            pylab.plot(xx,tmp)
            for fc, wdth, snr in peaksF:
                pylab.plot([fc,fc],[par.scanSNR,snr],'r')
                pylab.plot([fc-wdth/2,fc-wdth/2],[par.scanSNR/2,snr],'r')
                pylab.plot([fc+wdth/2,fc+wdth/2],[par.scanSNR/2,snr],'r')
            pylab.ylim([-20,int((5+4.34294*N.log(N.max(F)/noise))/5)*5])
            pylab.show()
        savePeaks += peaksF
    return savePeaks


def peakDetect(Fmean,look=False):
    Nfft = len(Fmean)
    # Better noise estimation
    noise, noiseStd = N.mean(Fmean), N.std(Fmean)
    cutoff=noise+3*noiseStd
    indx=N.where(Fmean<cutoff)[0]
    noise, noiseStd = N.mean(Fmean[indx]), N.std(Fmean[indx])
    # Detect peaks that are 3*std from mean
    cutoff=noise+3*noiseStd
    indx=N.where(Fmean>cutoff)[0]

    # Positions and first width
    found = []
    if len(indx)>0:
        last, first, mag = indx[0], indx[0], Fmean[indx[0]]
    
        for ii in indx[1:]:
            if ii>last+par.scanRes:
                snr = 4.34294*N.log(mag/(noise*(last-first+1))) # Power snr [DB]
                found=found+[[first,last,snr]]
                first=ii
                mag=0
            last=ii
            mag=mag+Fmean[ii]
        snr = 4.34294*N.log(mag/(noise*(last-first+1))) # Power snr in DB
        found=found+[[first,last,snr]]

    # Positions and more accurate width/snr
    found2 = []
    for first, last, void in found:
        if first!=last:
            subset = Fmean[first:last]
            max = N.max(subset)
            cutoff=noise*10**((par.scanSNR/2)/20)
            indx=N.where(subset>cutoff)[0]
            mag = N.sum(subset[indx])
            snr = 4.34294*N.log(mag/(noise*(len(indx)))) # Power snr in DB
        else:
            max = Fmean[first]
            snr = 4.34294*N.log(max/noise) # Power snr in DB
            indx=[0]
        # Remove first, last and middle 1/16 of the fft range 
        if first-Nfft/16<0 or Nfft-last<Nfft/16\
                or abs((first+last)/2-Nfft/2)<Nfft/32:
            snr=0
        if look and abs((first+last)/2.0-par.lookNfft*0.75)>par.lookNfft/8:
            snr=0
        if snr>par.scanSNR:
            found2=found2+[[first+N.min(indx),first+N.max(indx),snr]]

    return found2, noise, noiseStd
        

def main(radio,serverPort,verbose,calFreq):
    # Calibrate 
    radio.set_sample_rate(par.lookRate)
    radio.set_gain('auto')
    if calFreq!=None:
        calibrate(radio,par,calFreq*1e6,verbose)

    saved = {}
    while True:
        for ii in range(3):
            peekScanFreq = DB.clientSend(serverPort,['Peek?'])
            peekFoundFreq = peek(radio,par,verbose,peekScanFreq)
            DB.clientSend(serverPort,['PeekResults',peekFoundFreq])

        lookFreq = DB.clientSend(serverPort,['Look?'])
        for freq in lookFreq:
            found, freq, lfreq, fn = look(radio,par,freq,verbose)
            if found:
                DB.clientSend(serverPort,['LookResults',[lfreq,freq,fn]])
        
def start(device=0,serverPort=None,verbose=False,calFreq=None,snr=20.0):
    import multiprocessing
    par.scanSNR = snr
    radio=rtlsdr.RtlSdr(device_index=device)
    proc = multiprocessing.Process(target=main,args=(radio,serverPort,verbose,calFreq))
    proc.start()





